@@ -3,10 +3,12 @@ Verilog Codegen
 """
 
 import itertools
-import typing
+import logging
+import warnings
+from typing import Generator, cast
 
 from python2verilog.optimizer.optimizer import backwards_replace
-from python2verilog.utils.string import Lines
+from python2verilog.utils.lines import Lines
 
 from ... import ir
 from ...utils.assertions import assert_typed_dict, get_typed, get_typed_list
@@ -18,29 +20,27 @@ class CodeGen:
     Code Generator for Verilog
     """
 
-    def __init__(self, root: ir.Vertex, context: ir.Context):
+    def __init__(self, root: ir.Node, context: ir.Context):
         """ "
         Builds tree from Graph IR
         """
-        get_typed(root, ir.Vertex)
+        get_typed(root, ir.Node)
         get_typed(context, ir.Context)
         self.context = context
-
-        self.context.output_vars = [
-            ir.Var(str(i)) for i in range(len(self.context.output_types))
-        ]
-
         root_case = CaseBuilder(root, context).case
+        logging.debug(
+            f"{self.__class__.__name__} "
+            f"{[case.condition.ver_name for case in root_case.case_items]}"  # type: ignore
+        )
 
         for item in root_case.case_items:
-            self.context.add_state(item.condition.to_string())
+            self.context.add_state_weak(
+                item.condition.to_string()
+            )  # change to not weak
 
-        assert isinstance(context.ready_state, str)
-        self.context.add_state_weak(context.ready_state)
-
-        self.context.add_global_var(
-            ir.Var("state", initial_value=self.context.ready_state)
-        )
+        assert isinstance(context.done_state, ir.State)
+        self.context.add_state_weak(str(context.done_state))
+        self.context.add_state_weak(str(context.idle_state))
 
         self._module = CodeGen.__new_module(root_case, self.context)
 
@@ -51,7 +51,7 @@ class CodeGen:
 
         Requires context for I/O and declarations
         """
-        assert isinstance(root, ver.Statement)
+        assert isinstance(root, ver.Case)
         assert isinstance(context, ir.Context)
 
         inputs = []
@@ -62,22 +62,52 @@ class CodeGen:
         for var in context.output_vars:
             outputs.append(var.ver_name)
 
+        def make_debug_display(context: ir.Context):
+            """
+            Creates a display statement for all signals
+
+            $display("%0d, ...", ...);
+            """
+            vars_: list[str] = []
+            vars_ += map(
+                lambda x: x.ver_name, context.signals.instance_specific_values()
+            )
+            vars_ += map(lambda x: x.py_name, context.input_vars)
+            vars_ += map(lambda x: x.ver_name, context.input_vars)
+            vars_ += map(lambda x: x.ver_name, context.output_vars)
+            vars_ += map(lambda x: x.ver_name, context.global_vars)
+            str_ = f'$display("{context.name},%s,'
+            str_ += "=%0d,".join(vars_) + '=%0d", '
+            str_ += f"{context.state_var.ver_name}.name, "
+            str_ += ", ".join(vars_)
+            str_ += ");"
+            return str_
+
+        def create_instance_zeroed_signals() -> Generator[ver.Statement, None, None]:
+            """
+            Instance signals that should always be set to zero be default
+            """
+            for instance in context.instances.values():
+                yield ver.NonBlockingSubsitution(instance.signals.ready, ir.UInt(0))
+                yield ver.NonBlockingSubsitution(instance.signals.start, ir.UInt(0))
+
         always = ver.PosedgeSyncAlways(
-            context.clock_signal,
+            context.signals.clock,
             body=[
-                ver.NonBlockingSubsitution(context.done_signal, ir.UInt(0)),
+                ver.Statement("`ifdef DEBUG"),
+                ver.Statement(make_debug_display(context)),
+                ver.Statement("`endif"),
+                ver.NonBlockingSubsitution(context.signals.done, ir.UInt(0)),
+                ver.Statement(),
+            ]
+            + list(create_instance_zeroed_signals())
+            + [
                 ver.Statement(),
                 ver.IfElse(
-                    context.ready_signal,
-                    typing.cast(
+                    context.signals.ready,
+                    cast(
                         list[ver.Statement],
-                        [
-                            ver.NonBlockingSubsitution(out, ir.Int(0))
-                            for out in context.output_vars
-                        ]
-                        + [
-                            ver.NonBlockingSubsitution(context.valid_signal, ir.UInt(0))
-                        ],
+                        [ver.NonBlockingSubsitution(context.signals.valid, ir.UInt(0))],
                     ),
                     [],
                 ),
@@ -86,11 +116,11 @@ class CodeGen:
                 ver.Statement(),
                 ver.Statement(comment="Start signal takes precedence over reset"),
                 ver.IfElse(
-                    context.reset_signal,
+                    context.signals.reset,
                     then_body=[
                         ver.NonBlockingSubsitution(
                             lvalue=context.state_var,
-                            rvalue=ir.Expression(context.ready_state),
+                            rvalue=context.idle_state,
                         ),
                     ],
                     else_body=[],
@@ -105,18 +135,72 @@ class CodeGen:
             ver.Statement(comment="Global variables"),
         ]
 
-        body += [ver.Declaration(v, reg=True, signed=True) for v in context.global_vars]
+        body += [
+            ver.Declaration(v.ver_name, reg=True, signed=True)
+            for v in context.global_vars
+        ]
 
         body += [
             ver.Declaration(var.ver_name, reg=True, signed=True)
             for var in context.input_vars
         ]
 
-        body.append(ver.Statement())
+        for instance in context.instances.values():
+            body.append(
+                ver.Statement(
+                    comment="================ Function Instance ================"
+                )
+            )
+            module = context.namespace[instance.module_name]
+            defaults = {
+                module.signals.valid: instance.signals.valid,
+                module.signals.done: instance.signals.done,
+                module.signals.clock: context.signals.clock,
+                module.signals.start: instance.signals.start,
+                module.signals.reset: instance.signals.reset,
+                module.signals.ready: instance.signals.ready,
+            }
+            # defaults = dict(zip(module.signals.values(), instance.signals.values()))
+            for var in instance.inputs:
+                body.append(ver.Declaration(name=var.ver_name, reg=True))
+            for var in instance.outputs:
+                body.append(ver.Declaration(name=var.ver_name))
+            body.append(ver.Declaration(name=instance.signals.valid.ver_name, size=1))
+            body.append(ver.Declaration(name=instance.signals.done.ver_name, size=1))
+            body.append(
+                ver.Declaration(name=instance.signals.start.ver_name, size=1, reg=True)
+            )
+            body.append(
+                ver.Declaration(name=instance.signals.ready.ver_name, size=1, reg=True)
+            )
+            body.append(
+                ver.Instantiation(
+                    instance.module_name,
+                    instance.var.ver_name,
+                    {
+                        key.py_name: value.ver_name
+                        for key, value in zip(
+                            module.input_vars,
+                            instance.inputs,
+                        )
+                    }
+                    | {
+                        key.ver_name: value.ver_name
+                        for key, value in zip(
+                            module.output_vars,
+                            instance.outputs,
+                        )
+                    }
+                    | {key.ver_name: value.ver_name for key, value in defaults.items()},
+                )
+            )
 
+        body.append(ver.Statement(comment="Core"))
         body.append(always)
 
-        state_vars = {key: ir.UInt(index) for index, key in enumerate(context.states)}
+        state_vars = {
+            key: ir.UInt(index) for index, key in enumerate(sorted(context.states))
+        }
 
         python_test_code = Lines()
         for case in context.test_cases:
@@ -127,7 +211,7 @@ class CodeGen:
             outputs=outputs,
             body=body,
             localparams=state_vars,
-            header_comment=Lines(
+            header=Lines(
                 f"/*\n\n# Python Function\n{context.py_string}\n\n"
                 f"# Test Cases\n{python_test_code}\n*/\n\n"
             ),
@@ -144,39 +228,51 @@ class CodeGen:
             ...
         end
         """
-        # The first case can be included here
-        mapping = {
-            ir.Expression(var.ver_name): ir.Expression(var.py_name)
-            for var in context.input_vars
-        }
-
         then_body: list[ver.Statement] = []
-        for var in context.input_vars:
-            then_body.append(
-                ver.NonBlockingSubsitution(
-                    ir.Expression(var.ver_name), ir.Expression(var.py_name)
+        if context.optimization_level > 0:
+            # The first case can be included here
+            # Known as Quick Start
+            mapping = {
+                ir.Expression(var.ver_name): ir.Expression(var.py_name)
+                for var in context.input_vars
+            }
+
+            for var in context.input_vars:
+                then_body.append(
+                    ver.NonBlockingSubsitution(
+                        ir.Expression(var.ver_name), ir.Expression(var.py_name)
+                    )
                 )
+
+            stmt_stack: list[ver.Statement] = []  # backwards replace using dfs
+            for item in root.case_items:
+                if item.condition == context.entry_state:
+                    stmt_stack += item.statements
+                    then_body += item.statements
+                    root.case_items.remove(item)
+                    break
+
+            while stmt_stack:
+                stmt = stmt_stack.pop()
+                if isinstance(stmt, ver.NonBlockingSubsitution):
+                    stmt.rvalue = backwards_replace(stmt.rvalue, mapping)
+                elif isinstance(stmt, ver.IfElse):
+                    stmt.condition = backwards_replace(stmt.condition, mapping)
+                    stmt_stack += stmt.then_body
+                    stmt_stack += stmt.else_body
+                else:
+                    raise TypeError(f"Unexpected {type(stmt)} {stmt}")
+        else:
+            for var in context.input_vars:
+                then_body.append(
+                    ver.NonBlockingSubsitution(
+                        ir.Expression(var.ver_name), ir.Expression(var.py_name)
+                    )
+                )
+            then_body.append(
+                ver.NonBlockingSubsitution(context.state_var, context.entry_state)
             )
 
-        stmt_stack: list[ver.Statement] = []  # backwards replace using dfs
-        for item in root.case_items:
-            if item.condition == ir.Expression(context.entry_state):
-                stmt_stack += item.statements
-                then_body += item.statements
-                root.case_items.remove(item)
-                break
-
-        while stmt_stack:
-            stmt = stmt_stack.pop()
-            if isinstance(stmt, ver.NonBlockingSubsitution):
-                stmt.rvalue = backwards_replace(stmt.rvalue, mapping)
-            elif isinstance(stmt, ver.IfElse):
-                stmt.condition = backwards_replace(stmt.condition, mapping)
-                stmt_stack += stmt.then_body
-                stmt_stack += stmt.else_body
-            else:
-                raise TypeError(f"Unexpected {type(stmt)} {stmt}")
-
         if_else = ver.IfElse(
             ir.Expression("_start"),
             then_body,
@@ -185,16 +281,17 @@ class CodeGen:
                     comment="If ready or not valid, then continue computation"
                 ),
                 ver.IfElse(
-                    ir.BinOp(
-                        context.ready_signal,
+                    ir.UBinOp(
+                        context.signals.ready,
                         "||",
-                        ir.UnaryOp("!", context.valid_signal),
+                        ir.UnaryOp("!", context.signals.valid),
                     ),
                     [root],
                     [],
                 ),
             ],
         )
+        # logging.debug(f"make start if else {if_else}")
         return [if_else]
 
     @property
@@ -215,7 +312,7 @@ class CodeGen:
         """
         Get Verilog module as string
         """
-        return str(self.get_module_lines())
+        return self.get_module_lines().to_string()
 
     def get_testbench(self, random_wait: bool = False):
         """
@@ -226,11 +323,14 @@ class CodeGen:
         :param random_wait: whether or not to have random _wait signal in the while loop
         """
         if len(self.context.input_vars) == 0:
-            raise RuntimeError(f"Input var names not deduced for {self.context.name}")
+            raise RuntimeError(
+                f"Input var names not deduced for {self.context.name}, "
+                f"use {self.context.name} in your Python code"
+            )
         if len(self.context.output_vars) == 0:
             raise RuntimeError(
-                f"Output var types not deduced for {self.context.name}, \
-                    types are {self.context.output_types}"
+                f"Output var types not deduced for {self.context.name}, "
+                f"use {self.context.name} in your Python code"
             )
 
         def make_display_stmt():
@@ -243,7 +343,7 @@ class CodeGen:
             string += "%0d, " * (len(self.context.output_vars) - 1)
             string += '%0d", _valid, _ready'
             for var in self.context.output_vars:
-                string += f", {var}"
+                string += f", {var.ver_name}"
             string += ");"
             return ver.Statement(literal=string)
 
@@ -272,13 +372,13 @@ class CodeGen:
         setups.append(ver.Statement(literal="always #5 _clock = !_clock;"))
 
         initial_body: list[ver.Statement | ver.While] = []
-        initial_body.append(ver.BlockingSub(self.context.clock_signal, ir.UInt(0)))
-        initial_body.append(ver.BlockingSub(self.context.start_signal, ir.UInt(0)))
-        initial_body.append(ver.BlockingSub(self.context.ready_signal, ir.UInt(1)))
-        initial_body.append(ver.BlockingSub(self.context.reset_signal, ir.UInt(1)))
+        initial_body.append(ver.BlockingSub(self.context.signals.clock, ir.UInt(0)))
+        initial_body.append(ver.BlockingSub(self.context.signals.start, ir.UInt(0)))
+        initial_body.append(ver.BlockingSub(self.context.signals.ready, ir.UInt(1)))
+        initial_body.append(ver.BlockingSub(self.context.signals.reset, ir.UInt(1)))
 
-        initial_body.append(ver.AtNegedgeStatement(self.context.clock_signal))
-        initial_body.append(ver.BlockingSub(self.context.reset_signal, ir.UInt(0)))
+        initial_body.append(ver.AtNegedgeStatement(self.context.signals.clock))
+        initial_body.append(ver.BlockingSub(self.context.signals.reset, ir.UInt(0)))
         initial_body.append(ver.Statement())
 
         for i, test_case in enumerate(self.context.test_cases):
@@ -295,11 +395,11 @@ class CodeGen:
                         ir.Expression(var.py_name), ir.Int(int(test_case[i]))
                     )
                 )
-            initial_body.append(ver.BlockingSub(self.context.start_signal, ir.UInt(1)))
+            initial_body.append(ver.BlockingSub(self.context.signals.start, ir.UInt(1)))
 
             # Post-start
             initial_body.append(ver.Statement())
-            initial_body.append(ver.AtNegedgeStatement(self.context.clock_signal))
+            initial_body.append(ver.AtNegedgeStatement(self.context.signals.clock))
             for i, var in enumerate(self.context.input_vars):
                 initial_body.append(
                     ver.BlockingSub(
@@ -308,7 +408,7 @@ class CodeGen:
                         comment="only need inputs when start is set",
                     )
                 )
-            initial_body.append(ver.BlockingSub(self.context.start_signal, ir.UInt(0)))
+            initial_body.append(ver.BlockingSub(self.context.signals.start, ir.UInt(0)))
             initial_body.append(ver.Statement())
 
             # While loop waitng for ready signal
@@ -328,15 +428,25 @@ class CodeGen:
                     else_body=[],
                 )
             )
-            while_body.append(ver.AtNegedgeStatement(self.context.clock_signal))
+            while_body.append(ver.AtNegedgeStatement(self.context.signals.clock))
 
             initial_body.append(
                 ver.While(
-                    condition=ir.UnaryOp("!", self.context.done_signal),
+                    condition=ir.UBinOp(
+                        ir.UnaryOp("!", self.context.signals.done),
+                        "||",
+                        ir.UnaryOp("!", self.context.signals.ready),
+                    ),
                     body=while_body,
                 )
             )
-            initial_body.append(make_display_stmt())
+            initial_body.append(
+                ver.IfElse(
+                    condition=ir.Expression("_ready"),
+                    then_body=[make_display_stmt()],
+                    else_body=[],
+                )
+            )
             initial_body.append(ver.Statement())
 
         initial_body.append(ver.Statement(literal="$finish;"))
@@ -345,7 +455,7 @@ class CodeGen:
 
         if self.context:
             module = ver.Module(
-                f"{self.context.name}_tb",
+                self.context.testbench_name,
                 [],
                 [],
                 body=setups + [initial_loop],
@@ -364,7 +474,7 @@ class CodeGen:
         """
         New testbench as str
         """
-        return str(self.get_testbench_lines(random_wait=random_wait))
+        return self.get_testbench_lines(random_wait=random_wait).to_string()
 
 
 class CaseBuilder:
@@ -372,57 +482,77 @@ class CaseBuilder:
     Creates a case statement for the IR Graph
     """
 
-    def __init__(self, root: ir.Vertex, context: ir.Context):
+    def __init__(self, root: ir.Node, context: ir.Context):
         # Member Vars
         self.visited: set[str] = set()
         self.context = context
-        self.case = ver.Case(expression=ir.Expression("_state"), case_items=[])
-        self.added_ready_node = False
+        self.case = ver.Case(expression=context.state_var, case_items=[])
 
         # Member Funcs
         instance = itertools.count()
         self.next_unique = lambda: next(instance)
 
         # Work
+        logging.debug(f"{self.__class__.__name__} {root.unique_id} {root}")
         self.case.case_items.append(self.new_caseitem(root))
-        if not self.added_ready_node:
+        has_done = False
+        for case in self.case.case_items:
+            if case.condition == context.done_state:
+                has_done = True
+        if not has_done:
             self.case.case_items.append(
                 ver.CaseItem(
-                    ir.Expression(context.ready_state),
-                    statements=[
-                        ver.NonBlockingSubsitution(
-                            lvalue=ir.State("_ready"), rvalue=ir.UInt(1)
-                        )
-                    ],
+                    condition=context.done_state,
+                    statements=[self.create_quick_done(context)],
                 )
             )
+        self.case.case_items = list(reversed(self.case.case_items))
 
-    def new_caseitem(self, root: ir.Vertex):
+    @staticmethod
+    def create_quick_done(context: ir.Context) -> ver.IfElse:
+        """
+        if ready:
+            done = 1
+            state = idle
+        else:
+            state = done
+        """
+        return ver.IfElse(
+            condition=ir.BinOp(
+                ir.UnaryOp("!", context.signals.valid), "&&", context.signals.ready
+            ),
+            # condition=context.signals.ready,
+            then_body=[
+                ver.NonBlockingSubsitution(context.signals.done, ir.UInt(1)),
+                ver.NonBlockingSubsitution(context.state_var, context.idle_state),
+            ],
+            else_body=[
+                ver.NonBlockingSubsitution(context.state_var, context.done_state),
+            ],
+        )
+
+    def new_caseitem(self, root: ir.Node):
         """
         Creates a new case item with the root's unique id as identifier
         """
         stmts = self.do_vertex(root)
-        item = ver.CaseItem(condition=ir.Expression(root.unique_id), statements=stmts)
+        logging.debug(f"new caseitem {root.unique_id}")
+        item = ver.CaseItem(condition=ir.State(root.unique_id), statements=stmts)
 
         return item
 
-    def do_vertex(self, vertex: ir.Vertex):
+    def do_vertex(self, vertex: ir.Node):
         """
         Processes a node
         """
-        assert isinstance(vertex, ir.Vertex)
+
+        assert isinstance(vertex, ir.Node), str(vertex)
         self.visited.add(vertex.unique_id)
 
         stmts: list[ver.Statement] = []
 
         if isinstance(vertex, ir.DoneNode):
-            stmts += [
-                ver.NonBlockingSubsitution(self.context.done_signal, ir.UInt(1)),
-                ver.NonBlockingSubsitution(
-                    self.case.condition, ir.Expression(self.context.ready_state)
-                ),
-            ]
-            self.added_ready_node = True
+            stmts.append(self.create_quick_done(self.context))
 
         elif isinstance(vertex, ir.AssignNode):
             stmts.append(ver.NonBlockingSubsitution(vertex.lvalue, vertex.rvalue))
@@ -440,21 +570,23 @@ class CaseBuilder:
             )
 
         elif isinstance(vertex, ir.YieldNode):
-            outputs = [
+            outputs: list[ver.Statement] = []
+            outputs += [
                 ver.NonBlockingSubsitution(var, expr)
                 for var, expr in zip(self.context.output_vars, vertex.stmts)
-            ] + [ver.NonBlockingSubsitution(self.context.valid_signal, ir.UInt(1))]
-            state_change = []
+            ]
+            outputs += [
+                ver.NonBlockingSubsitution(self.context.signals.valid, ir.UInt(1))
+            ]
+            state_change: list[ver.Statement] = []
 
             if isinstance(vertex.optimal_child.optimal_child, ir.DoneNode):
-                outputs.append(
-                    ver.NonBlockingSubsitution(self.context.done_signal, ir.UInt(1))
-                )
+                outputs.append(self.create_quick_done(self.context))
 
             state_change.append(
                 ver.NonBlockingSubsitution(
                     self.context.state_var,
-                    ir.Expression(vertex.optimal_child.optimal_child.unique_id),
+                    ir.State(vertex.optimal_child.optimal_child.unique_id),
                 )
             )
             if vertex.optimal_child.optimal_child.unique_id not in self.visited:
@@ -462,7 +594,6 @@ class CaseBuilder:
                     self.new_caseitem(vertex.optimal_child.optimal_child)
                 )
 
-            # stmts += outputs + self.do_edge(vertex.optimal_child)
             stmts += outputs + state_change
 
         else:
@@ -481,7 +612,7 @@ class CaseBuilder:
                 self.case.case_items.append(self.new_caseitem(edge.optimal_child))
             return [
                 ver.NonBlockingSubsitution(
-                    self.context.state_var, ir.Expression(edge.optimal_child.unique_id)
+                    self.context.state_var, ir.State(edge.optimal_child.unique_id)
                 )
             ]
         raise RuntimeError(f"{type(edge)}")
